/*
   MRTG 2.8.12  -- Rateup
   *********************

   Rateup is a fast add-on to the great MRTG Traffic monitor.  It makes
   the database file updates much faster, and creates the graphic image
   files, ready for processing by PPMTOGIF.  It also reduces memory
   requirements by a factor of 10, and increases the speed of updates
   by a factor of at least 10.  This makes it feasible to run mrtg
   every 5 minutes.

   rateup attempts to compensate for missed updates by repeating the last
   sample, and also tries to catch bad update times.  The .log file stores
   real history every five minutes for 31 hours, then 'compresses' the 
   history into 30 minute samples for a week, then 2-hour samples for
   31 days, then daily samples for two years.  This ensures that the
   log files don't grow in size.

   The log files are a slightly different format, but convert.perl
   will fix that for you.

   Enjoy!
   Dave Rand
   dlr@bungi.com

   04/26/99 - There was some compilation bug under Watcom 10.6
   which was fixed when recompiled with VC++ 6.0
   Alexandre Steinberg
   steinberg@base.com.br
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <ctype.h>
#include <errno.h>
#define GFORM_GD gdImagePng
/* BSDI 2.0.1 does not have malloc.h */
#if !defined(bsdi) && !defined(__FreeBSD__) && !defined(__OpenBSD__)
#include <malloc.h>
#endif

#include <gd.h>
#include <gdfonts.h>

#include <tagr.h>

static char htmlf[1024];
// static char *VERSION = "2.8.12";
static char *program, *router, *routerpath;
static int histvalid;
static int dorelpercent = 0;
static int transparent = 0;
static int unknaszero = 0;

static time_t NOW;
static char *host_ip;
static int host_speed;

static char *short_si[] = {"", "k", "M", "G", "T"};
static char *longup = NULL;
static char *shortup = NULL;
static char weekformat = 'W';		/* strftime() fmt char for week #  */

#define DAY_COUNT (600)		/* 400 samples is 33.33 hours */
#define DAY_SAMPLE (5*60)	/* Sample every 5 minutes */
#define WEEK_COUNT (600)	/* 400 samples is 8.33 days */
#define WEEK_SAMPLE (30*60)	/* Sample every 30 minutes */
#define MONTH_COUNT (600)	/* 400 samples is 33.33 days */
#define MONTH_SAMPLE (2*60*60)	/* Sample every 2 hours */
#define YEAR_COUNT  (2 * 366)	/* 1 sample / day, 366 days, 2 years */
#define YEAR_SAMPLE (24*60*60)	/* Sample every 24 hours */

/* One 'rounding error' per sample period, so add 4 to total and for 
   good mesure we take 10 :-) */
#define MAX_HISTORY (DAY_COUNT+WEEK_COUNT+MONTH_COUNT+YEAR_COUNT+10)

/* These are the colors used for the variouse parts of the graph */
/* the format is Red,Green,Blue */

#define c_blank 245,245,245
#define c_light 194,194,194
#define c_dark 100,100,100
#define c_major 255,0,0
#define c_in    0,235,12
#define c_out   0,94,255
#define c_grid  0,0,0
#define c_inm   0,166,33
#define c_outm  255,0,255
#define c_outp  239,159,79

static int col_in[3];
static int col_out[3];
static int col_inm[3];
static int col_outm[3];
static int col_outp[3];

static long kilo = (long) 1000;
static char *kMG = (char *) NULL;
static char *kMGcopy = (char *) NULL;	/* JML bugfix */
static int kMGnumber = 0;

#define MAXL	200		/* Maximum length of last in & out fields */


struct HISTORY {
    time_t time;
    unsigned long in, out, percent, inmax, outmax;
} *history;

static int Mh;

struct LAST {
    time_t time;
    char in[MAXL], out[MAXL];
} last;

#if !defined(__WATCOMC__) && !defined(WIN32)
#define max(a,b) ((a) > (b) ? (a) : (b))
#define min(a,b) ((a) < (b) ? (a) : (b))
#endif

/*

   notes about the NEXT macro ....

   * position n to the entry in the history array so that NOW is between
   history[n] and history[n+1]

   * calculate the interval according to steptime and position of 
   now within the history array.

   for debuging 

   fprintf (stderr,"NOW: %8lu  ST: %4lu  N: %4u HTN: %8lu HTN+1: %8lu IV: %6.0f\n", \
   now,steptime,n,history[n].time, history[n+1].time, interval);\

 */

#define NEXT(steptime) \
  inmax = outmax = avc = 0; \
  inr = outr = 0.0;\
  nextnow = now - steptime;\
  if (now==history[n].time && n<histvalid && nextnow==history[n+1].time) { \
    inr = (double) history[n].in;\
    outr = (double) history[n].out;\
    inmax = history[n].inmax;\
    outmax = history[n].outmax;\
    n++;\
  } else {\
    if (now > history[n].time) {\
     fprintf(stderr,"ERROR: trying to read ahead of the available data\n"); \
    } else {\
      while (now <= history[n+1].time && n < histvalid){n++;}\
      do {\
	if (now >= history[n].time) {\
	  if (nextnow <= history[n+1].time) {\
	    interval = history[n].time - history[n+1].time;\
	  } else {\
	    interval = history[n].time - nextnow;\
	  }\
	} else {\
	  if (nextnow > history[n+1].time) {\
	    interval = steptime;\
	  } else {\
	    interval = now - history[n+1].time;\
	  }\
	}\
	inr  += (double) history[n].in * interval;\
	outr += (double) history[n].out * interval;\
	avc += interval;\
	inmax =  max(inmax, history[n].inmax);\
	outmax = max(outmax,history[n].outmax);\
	if (nextnow <= history[n+1].time) n++; else break;\
      } while (n < histvalid && nextnow < history[n].time);\
\
      if (avc != steptime) {\
	fprintf(stderr,"ERROR: StepTime does not match Avc %8lu\n", avc);\
      }\
\
      inr /= avc; outr /= avc;\
    }\
  }

struct image_descr {
	unsigned long maxx;
	unsigned long maxy;
	double xscale, yscale;
	unsigned long growright, step, bits;
	int ytics;
	double yticsf;
	int peak;

	const char *maxin_name;
	const char *avgin_name;
	const char *curin_name;
	const char *maxout_name;
	const char *avgout_name;
	const char *curout_name;
};

int
image (const char *file, long maxvi, long maxvo, const struct image_descr *id,
       pp_tab_t **tab)
{
    FILE *fo;
    int i, x, y, n, type;

    long maxv;
    unsigned long origmaxvi, origmaxvo, maxs, avc, inmax, outmax;
    double inr, outr, muli, interval;
    time_t now, onow, nextnow;
    struct tm *tm;
    char *graph_label[600];
    char ylab[30];
    /* scaling helpers */
    long maxv_q;
    unsigned long valsamp, maxin, maxout, digits, maxpercent = 0;
    unsigned long sca_ten, sca_hun, nmax_q;
    double avin, avout, latestout = 0.0, latestin = 0.0, nmax, avpercent = 0.0,
      latestpercent = 0.0;
    double nex_ten, nex_hun, nex_rnd;
    double sca_max_q, dummy;
    double percent;
    char *short_si_out;


    struct HISTORY *lhist;
    /* ################################################# */
    /* Some general definitions for the graph generation */
#define XSIZE (unsigned long)((id->maxx*id->xscale)+100+dorelpercent*30)
#define YSIZE (unsigned long)((id->maxy*id->yscale)+35)
    /* position the graph */
#define ytr(y) (unsigned long)(id->maxy*id->yscale+14-((y)*id->yscale))
    /* translate x/y coord into graph coord */
#define xtr(x) (unsigned long)(id->growright ? (id->maxx*id->xscale+81-((x)*id->xscale)) : (81+((x)*id->xscale)))
    /* ################################################# */

    /* GD LIB declarations */
    /* Declare the image */
    gdImagePtr graph, brush_out, brush_outm, brush_outp;
    /* Declare color indexes */
    int i_light, i_dark, i_blank, i_major, i_in, i_out, i_grid, i_inm,
      i_outm;
    int i_outp, i_outpg;
    /* Dotted style */
    int styleDotted[3];

    /* multiplicator for bits/bytes */
    muli = 7 * id->bits + 1;
    maxvi *= muli;
    maxvo *= muli;
    origmaxvi = labs (maxvi);
    origmaxvo = labs (maxvo);
    maxv = max (maxvi, maxvo);
    if (id->step > MONTH_SAMPLE) {
	type = 4;
	now = (long) (NOW / YEAR_SAMPLE) * YEAR_SAMPLE;
    } else if (id->step > WEEK_SAMPLE) {
	type = 3;
	now = (long) (NOW / MONTH_SAMPLE) * MONTH_SAMPLE;
    } else if (id->step > DAY_SAMPLE) {
	type = 2;
	now = (long) (NOW / WEEK_SAMPLE) * WEEK_SAMPLE;
    } else {
	type = 1;
	now = (long) (NOW / DAY_SAMPLE) * DAY_SAMPLE;
    }
    if ((lhist = calloc (1, sizeof (struct HISTORY) * id->maxx)) == NULL) {
	fprintf (stderr, "ERROR: Out of memory in graph creation\n");
	return (1);
    }
    onow = now;
    avin = avout = 0.0;
    inmax = outmax = maxin = maxout = 0;
    valsamp = 0;
    for (maxs = 0, n = 0, x = 0; x < id->maxx; x++) {
	NEXT (id->step);
	/*scale with muli */
	inr *= muli;
	outr *= muli;
	inmax *= muli;
	outmax *= muli;
	/* ignore times when we have not sampled */
	if (inr > 0 || outr > 0)
	    valsamp++;
	if (x == 0) {
	    latestin = inr;
	    latestout = outr;
	    if (outr) {
		latestpercent = inr * (double) 100.0 / outr;
	    }
	}
	avin += inr;
	avout += outr;
	if (id->peak) {
	    maxin = max (maxin, inmax);
	    maxout = max (maxout, outmax);
	    maxs = max (maxs, inmax);
	    maxs = max (maxs, outmax);
	} else {
	    maxin = max (maxin, inr);
	    maxout = max (maxout, outr);
	    maxs = max (maxs, inr);
	    maxs = max (maxs, outr);
	}
	if (dorelpercent && outr) {
	    dummy = (double) 100.0 * inr / outr;
	    maxpercent = max (dummy, maxpercent);
	}
	now -= id->step;
    }
    if (dorelpercent) {
	if (avout && valsamp) {
	    avpercent = (double) 100. *avin / avout;
	} else {
	    avpercent = 0;
	}
    }
    if (valsamp) {
	avin /= valsamp;
	avout /= valsamp;
    }

    add_numeric_value(tab, id->maxin_name, maxin / muli / 1024.0);
    add_numeric_value(tab, id->avgin_name, avin / muli / 1024.0);
    add_numeric_value(tab, id->curin_name, latestin / muli / 1024.0);
    add_numeric_value(tab, id->maxout_name, maxout / muli / 1024.0);
    add_numeric_value(tab, id->avgout_name, avout / muli / 1024.0);
    add_numeric_value(tab, id->curout_name, latestout / muli / 1024.0);

    if (maxv < 0 || maxv < maxs) {
	maxv = maxs;
    }
    now = onow;

    if (maxv <= 0)
	maxv = 1;

    if (kMG) {
	char *string = kMG;
	kMGnumber = 0;
	while (strchr (string, ',')) {
	    string = strchr (string, ',') + 1;
	    kMGnumber++;
	}
    } else {
	kMGnumber = 4;
    }

    /* mangle the 0.25*maxv value so, that we get a number with either */
    /* one or two digits != 0 and these digits should be at the  */
    /* start of the number ... */

    /* the ceil compensates for rounding with small numbers */
    maxv_q = ceil ((double) maxv / (double) id->ytics);	/* int */

    digits = 0;
    {
	double number = (double) maxv_q;
/*
   while (number/(double) kilo >= (double)kilo && digits<kMGnumber*3) {
 */
	while (number >= (double) kilo && digits < kMGnumber * 3) {
	    number /= (double) kilo;
	    digits += 3;
	}
	sca_max_q = (double) ((int) (((double) 100. * (double) number) /
	      (pow ((double) 10., (double) (int) (log10 ((double) number))))
				 + (double) 9.999) / (int) 10) / (double) 10
	    * (pow ((double) 10., (double) (int) (log10 ((double) number))));
    }

    if (kMG) {
	int count = (int) digits;
	short_si_out = kMGcopy;	/* JML bugfix */
	strcpy (kMGcopy, kMG);	/* JML bugfix */
	while (count >= 3) {
	    count -= 3;
	    short_si_out = strchr (short_si_out, ',') + 1;
	}
	if (strchr (short_si_out, ',')) {
	    *((char *) strchr (short_si_out, ',')) = '\0';
	}
    } else {
	short_si_out = short_si[min (digits / 3, kMGnumber)];
    }

    if (maxv_q) {
	digits = log10 ((double) maxv_q);
    } else {
	digits = 0;
    }

    sca_ten = maxv_q / pow (10.0, (double) digits);
    sca_hun = maxv_q / pow (10.0, (double) digits - 1);
    nex_rnd = (sca_hun) * pow (10.0, (double) digits - 1);
    nex_ten = (sca_ten + 1) * pow (10.0, (double) digits);
    nex_hun = (sca_hun + 1) * pow (10.0, (double) digits - 1);

    if (nex_ten <= (1.1 * maxv_q)) {
	nmax_q = nex_ten;
    } else if (maxv_q == nex_rnd) {
	nmax_q = nex_rnd;
    } else {
	nmax_q = nex_hun;
    }

    sca_max_q = nmax_q / (pow (10.0, (double) ((int) (digits / 3)) * 3));
    nmax = sca_max_q * id->ytics * (pow ((double) kilo, (double) ((int) (digits / 3))));

    if (nmax < 1.) {
	nmax = 1.;
    }
    for (n = 0, x = 0; x < id->maxx; x++) {
	lhist[x].time = 0;
	graph_label[x] = NULL;
	tm = localtime (&history[n].time);
	switch (type) {
	default:
	case 1:
	    if (tm->tm_min < 5) {
		lhist[x].time |= 1;
		if (tm->tm_hour == 0)
		    lhist[x].time |= 2;
	    }
	    if ((tm->tm_min < 5) && (tm->tm_hour % 2 == 0)) {
		if ((graph_label[x] = calloc (1, sizeof (char) * 4)) == NULL) {
		    fprintf (stderr, "ERROR: Out of memory in graph labeling\n");
		    return (1);
		} else {
		    sprintf (graph_label[x], "%i", tm->tm_hour);
		}
	    }
	    break;
	case 2:
	    if (tm->tm_min < 30 && tm->tm_hour == 0) {
		lhist[x].time |= 1;
		if (tm->tm_wday == 1)
		    lhist[x].time |= 2;
	    }
	    /* fprintf(stderr,"x: %i, min:%i, hour:%i day: %i\n",
	       x,tm->tm_min,tm->tm_hour,tm->tm_wday); */
	    if ((tm->tm_min < 30) &&
		(tm->tm_hour == 12)) {
		if ((graph_label[x] = calloc (1, sizeof (char) * 5)) == NULL) {
		    fprintf (stderr, "ERROR: Out of memory in graph labeling\n");
		    return (1);
		} else {
		    strftime (graph_label[x], 4, "%a", tm);
		}
	    }
	    break;
	case 3:
	    if (tm->tm_hour < 2) {
		if (tm->tm_wday == 1)
		    lhist[x].time |= 1;
		if (tm->tm_mday == 1)
		    lhist[x].time |= 2;
	    }
	    /* label goes to thursday noon */

	    if ((tm->tm_hour > 10) &&
		(tm->tm_hour <= 12) &&
		(tm->tm_wday == 4)) {
		if ((graph_label[x] = calloc (1, sizeof (char) * 16)) == NULL) {
		    fprintf (stderr, "ERROR: Out of memory in graph labeling\n");
		    return (1);
		} else {
		    char fmtbuf[10];
		    sprintf (fmtbuf, "Week %%%c", weekformat);
		    strftime (graph_label[x], 8, fmtbuf, tm);
		}
	    }
	    break;
	case 4:
	    if (tm->tm_mday == 1)
		lhist[x].time |= 1;
	    if (tm->tm_yday == 0)
		lhist[x].time |= 2;


	    if (tm->tm_mday == 15) {
		if ((graph_label[x] = calloc (1, sizeof (char) * 5)) == NULL) {
		    fprintf (stderr, "Rateup Error: Out of memory in graph labeling\n");
		    return (1);
		} else {
		    strftime (graph_label[x], 4, "%b", tm);
		}
	    }
	    break;
	}

	NEXT (id->step);
	/*scale with muli */
	inr *= muli;
	outr *= muli;
	inmax *= muli;
	outmax *= muli;

	y = ((double) inr / nmax) * id->maxy;
	if (y >= id->maxy)
	    y = id->maxy;
	lhist[x].in = y;
	y = ((double) outr / nmax) * id->maxy;
	if (y >= id->maxy)
	    y = id->maxy;
	lhist[x].out = y;
	y = ((double) inmax / nmax) * id->maxy;
	if (y >= id->maxy)
	    y = id->maxy;
	if (x && y < (lhist[x - 1].inmax)) {
	    y += (lhist[x - 1].inmax);
	    y /= 2;
	}
	lhist[x].inmax = y;
	y = ((double) outmax / nmax) * id->maxy;
	if (y >= id->maxy)
	    y = id->maxy;
	if (x && y < (lhist[x - 1].outmax)) {
	    y += (lhist[x - 1].outmax);
	    y /= 2;
	}
	lhist[x].outmax = y;

	if (dorelpercent) {
	    if (outr != (unsigned long) 0) {
		percent = (double) inr / (double) outr;
/*
   if (lhist[x].out != (unsigned long)0) {
   percent = (double)lhist[x].in/(double)lhist[x].out;
   if (lhist[x].in != (double)0.) {
   percent = (double)lhist[x].out/lhist[x].in;
 */
	    } else {
		percent = (double) 0.;
	    }
	    if (percent > (double) 1)
		percent = (double) 1.;
	    percent *= id->maxy;
	    lhist[x].percent = (unsigned long) percent;
	} else {
	    lhist[x].percent = (unsigned long) 0;
	}

	now -= id->step;
    }

    /* the graph is made ten pixels higher to acomodate the x labels */
    graph = gdImageCreate (XSIZE, YSIZE);
    brush_out = gdImageCreate (1, 2);
    brush_outm = gdImageCreate (1, 2);
    brush_outp = gdImageCreate (1, 2);

    /* the first color allocated will be the background color. */
    i_blank = gdImageColorAllocate (graph, c_blank);
    i_light = gdImageColorAllocate (graph, c_light);
    i_dark = gdImageColorAllocate (graph, c_dark);

    if (transparent) {
	gdImageColorTransparent (graph, i_blank);
    }
    gdImageInterlace (graph, 1);

    i_major = gdImageColorAllocate (graph, c_major);
    i_in = gdImageColorAllocate (graph, col_in[0], col_in[1], col_in[2]);
    i_out = gdImageColorAllocate (brush_out, col_out[0], col_out[1], col_out[2]);
    i_grid = gdImageColorAllocate (graph, c_grid);
    i_inm = gdImageColorAllocate (graph, col_inm[0], col_inm[1], col_inm[2]);
    i_outm = gdImageColorAllocate (brush_outm, col_outm[0], col_outm[1], col_outm[2]);
    i_outp = gdImageColorAllocate (brush_outp, col_outp[0], col_outp[1], col_outp[2]);
    i_outpg = gdImageColorAllocate (graph, col_outp[0], col_outp[1], col_outp[2]);

    /* draw the image border */
    gdImageLine (graph, 0, 0, XSIZE - 1, 0, i_light);
    gdImageLine (graph, 1, 1, XSIZE - 2, 1, i_light);
    gdImageLine (graph, 0, 0, 0, YSIZE - 1, i_light);
    gdImageLine (graph, 1, 1, 1, YSIZE - 2, i_light);
    gdImageLine (graph, XSIZE - 1, 0, XSIZE - 1, YSIZE - 1, i_dark);
    gdImageLine (graph, 0, YSIZE - 1, XSIZE - 1, YSIZE - 1, i_dark);
    gdImageLine (graph, XSIZE - 2, 1, XSIZE - 2, YSIZE - 2, i_dark);
    gdImageLine (graph, 1, YSIZE - 2, XSIZE - 2, YSIZE - 2, i_dark);


    /* draw the incoming traffic */
    for (x = 0; x < id->maxx; x++) {
	/* peak is always above average, we therefore only draw the upper part */
        if (id->peak)
	    gdImageLine (graph, xtr (x), ytr (lhist[x].in),
			 xtr (x), ytr (lhist[x].inmax), i_inm);
#ifdef INCOMING_UNFILLED
	gdImageLine (graph, xtr (x), ytr (lhist[x].in), xtr (x), ytr (lhist[x + 1].in), i_in);
#else
	gdImageLine (graph, xtr (x), ytr (0), xtr (x), ytr (lhist[x].in), i_in);
	/* draw the line a second time offset slightly. makes the graph
	 * look better if xscale is fractional */
	gdImageLine (graph, xtr (x + 0.5), ytr (0), xtr (x + 0.5), ytr (lhist[x].in), i_in);
#endif
    }

    /* draw the percentage */
/*
    if (dorelpercent) {
	gdImageSetBrush (graph, brush_outp);
	for (x = 0; x < maxx - 1; x++)
	    gdImageLine (graph, xtr (x), ytr (lhist[x].percent),
			 xtr (x + 1), ytr (lhist[x + 1].percent),
			 gdBrushed);
    }
*/
    /* draw the outgoing traffic */
    gdImageSetBrush (graph, brush_outm);
    if (id->peak)
	for (x = 0; x < id->maxx - 1; x++)
	    gdImageLine (graph, xtr (x), ytr (lhist[x].outmax),
			 xtr (x + 1), ytr (lhist[x + 1].outmax),
			 gdBrushed);
    gdImageSetBrush (graph, brush_out);
    for (x = 0; x < id->maxx - 1; x++)
	gdImageLine (graph, xtr (x), ytr (lhist[x].out),
		     xtr (x + 1), ytr (lhist[x + 1].out),
		     gdBrushed);

    /* draw the graph border */

    gdImageRectangle (graph,
		      xtr (0), ytr (0),
		      xtr (id->maxx), ytr (id->maxy), i_grid);

    /*create a dotted style for the grid lines */
    styleDotted[0] = i_grid;
    styleDotted[1] = gdTransparent;
    styleDotted[2] = gdTransparent;
    gdImageSetStyle (graph, styleDotted, 3);

    /* draw the horizontal grid */
    if ((longup == NULL) || (shortup == NULL)) {
	if (!id->bits) {
	    longup = "Bytes per Second";
	    shortup = "Bytes/s";
	} else {
	    longup = "Bits per Second";
	    shortup = "Bits/s";
	}
    }
    if (id->maxy < gdFontSmall->w * 16) {
	gdImageStringUp (graph, gdFontSmall, 8, ytr ((id->maxy - gdFontSmall->w * strlen (shortup)) / 2),
			 (unsigned char *) shortup, i_grid);
    } else {
	gdImageStringUp (graph, gdFontSmall, 8, ytr ((id->maxy - gdFontSmall->w * strlen (longup)) / 2),
			 (unsigned char *) longup, i_grid);
    }
    for (i = 0; i <= id->ytics; i++) {

	gdImageLine (graph, xtr (-2), ytr (i * id->maxy / id->ytics),
		     xtr (1), ytr (i * id->maxy / id->ytics), i_grid);

	gdImageLine (graph, xtr (id->maxx + 2), ytr (i * id->maxy / id->ytics),
		     xtr (id->maxx - 1), ytr (i * id->maxy / id->ytics), i_grid);

	gdImageLine (graph, xtr (0), ytr (i * id->maxy / id->ytics),
		     xtr (id->maxx), ytr (i * id->maxy / id->ytics), gdStyled);

/*
   sprintf(ylab,"%6.1f %s",sca_max_q*i,short_si[digits/3]);
 */
	sprintf (ylab, "%6.1f %s", sca_max_q * i * id->yticsf, short_si_out);
	gdImageString (graph, gdFontSmall, 23,
		       ytr (i * id->maxy / id->ytics + gdFontSmall->h / 2),
		       (unsigned char *) ylab, i_grid);

	if (dorelpercent) {
	    sprintf (ylab, "% 6.1f%%", (float) 25. * i);
	    gdImageString (graph, gdFontSmall,
			   77 + (id->maxx * id->xscale) + 1,
			   ytr (i * id->maxy / id->ytics + gdFontSmall->h / 2),
			   (unsigned char *) ylab, i_outpg);
	}
    }

    /* draw vertical grid and horizontal labels */
    for (x = 0; x < id->maxx; x++) {
	if (lhist[x].time) {
	    gdImageLine (graph, xtr (x), ytr (-2), xtr (x), ytr (1), i_grid);
	    gdImageLine (graph, xtr (x), ytr (0), xtr (x), ytr (id->maxy), gdStyled);
	}
	if (graph_label[x] != NULL) {
	    gdImageString (graph, gdFontSmall,
			   (xtr (x) - (strlen (graph_label[x]) *
				       gdFontSmall->w / 2)),
			   ytr (-4),
			   (unsigned char *) graph_label[x], i_grid);
	    free (graph_label[x]);
	}
	if (lhist[x].time & 2)
	    gdImageLine (graph, xtr (x), ytr (0), xtr (x), ytr (id->maxy), i_major);
    }

    /* draw line at peak In value in i_major color */
    /* only draw the line if it's within the graph ... */
    if (ytr (id->maxy) < ytr (((double) origmaxvi / nmax) * id->maxy)) {
	styleDotted[0] = i_major;
	gdImageSetStyle (graph, styleDotted, 3);
	gdImageLine (graph,
		     xtr (0), ytr (((double) origmaxvi / nmax) * id->maxy),
		     xtr (id->maxx), ytr (((double) origmaxvi / nmax) * id->maxy),
		     gdStyled);
    }
    /* draw line at peak Out value in i_major color */
    /* only draw the line if it's within the graph ... */
    if (ytr (id->maxy) < ytr (((double) origmaxvo / nmax) * id->maxy)) {
	styleDotted[0] = i_major;
	gdImageSetStyle (graph, styleDotted, 3);
	gdImageLine (graph,
		     xtr (0), ytr (((double) origmaxvo / nmax) * id->maxy),
		     xtr (id->maxx), ytr (((double) origmaxvo / nmax) * id->maxy),
		     gdStyled);
    }
    /* draw a red arrow a 0,0 */
    gdImageLine (graph, xtr (2), ytr (3), xtr (2), ytr (-3), i_major);
    gdImageLine (graph, xtr (1), ytr (3), xtr (1), ytr (-3), i_major);
    gdImageLine (graph, xtr (0), ytr (2), xtr (0), ytr (-2), i_major);
    gdImageLine (graph, xtr (-1), ytr (1), xtr (-1), ytr (-1), i_major);
    gdImageLine (graph, xtr (-2), ytr (1), xtr (-2), ytr (-1), i_major);
    gdImageLine (graph, xtr (-3), ytr (0), xtr (-3), ytr (0), i_major);


    if ((fo = fopen (file, "wb")) == NULL) {
	perror (program);
	fprintf (stderr, "Rateup Error: Can't open %s for write\n", file);
	exit (1);
    }
    GFORM_GD (graph, fo);
    fclose (fo);
    gdImageDestroy (graph);
    gdImageDestroy (brush_out);
    gdImageDestroy (brush_outm);
    gdImageDestroy (brush_outp);
    free (lhist);
    
    return 0;
    
}


double
diff (a, b)
     char *a, *b;
{
    char res[MAXL], *a1, *b1, *r1;
    int c, x, m;

    while (!isdigit (*a))
	a++;
    while (!isdigit (*b))
	b++;
    a1 = &a[strlen (a) - 1];
    m = max (strlen (a), strlen (b));
    r1 = &res[m + 1];
    for (b1 = res; b1 <= r1; b1++)
	*b1 = ' ';
    b1 = &b[strlen (b) - 1];
    r1[1] = 0;			/* Null terminate result */
    c = 0;
    for (x = 0; x < m; x++) {
	if (a1 >= a && b1 >= b) {
	    *r1 = ((*a1 - c) - *b1) + '0';
	} else if (a1 >= a) {
	    *r1 = (*a1 - c);
	} else {
	    *r1 = ('0' - *b1 - c) + '0';
	}
	if (*r1 < '0') {
	    *r1 += 10;
	    c = 1;
	} else {
	    c = 0;
	}
	a1--;
	b1--;
	r1--;
    }
    if (c) {
	r1 = &res[m + 1];
	for (x = 0; isdigit (*r1) && x < m; x++, r1--) {
	    *r1 = ('9' - *r1 + c) + '0';
	    if (*r1 > '9') {
		*r1 -= 10;
		c = 1;
	    } else {
		c = 0;
	    }
	}
	return (-atof (res));
    } else
	return (atof (res));
}

int
readhist (file)
     char *file;
{
    FILE *fi;
    int i, x, retcode = 0;
    char buf[256];
    struct HISTORY *hist;
    unsigned long rd[5];
    time_t cur;

    if ((fi = fopen (file, "r")) != NULL) {
	if (fscanf (fi, "%ld %s %s\n", &last.time, &last.in[0], &last.out[0]) != 3) {
	    fprintf (stderr, "Read Error: File %s lin 1\n", file);
	    retcode = 1;
	}
	cur = last.time;
	x = histvalid = 0;
	hist = history;
	while (!feof (fi)) {
	    fgets (buf, 256, fi);
	    if (sscanf (buf, "%lu %lu %lu %lu %lu",
			&rd[0], &rd[1], &rd[2], &rd[3], &rd[4]) < 5) {
		rd[3] = rd[1];
		rd[4] = rd[2];
	    }
	    for (i = 0; i <= 4; i++) {
		rd[i] = rd[i] < 0 ? 0 : rd[i];
	    }

	    hist->time = rd[0];
	    hist->in = rd[1];
	    hist->out = rd[2];
	    hist->inmax = rd[3];
	    hist->outmax = rd[4];
	    if (hist->inmax < hist->in)
		hist->inmax = hist->in;
	    if (hist->outmax < hist->out)
		hist->outmax = hist->out;
	    if (hist->time > cur) {
		//fprintf(stderr,"ERROR: %s found %s's log file was corrupt\n          or not in sorted order:\ntime: %lu.",program,router,(unsigned long)hist->time);
		retcode = 2;
		break;
	    }
	    cur = hist->time;
	    if (++x >= Mh) {
		struct HISTORY *lh;

		Mh += MAX_HISTORY;
		lh = realloc (history, (Mh + 1) * sizeof (struct HISTORY));
		if (lh == NULL) {
		    //fprintf(stderr,"Rateup WARNING: (pay attention to this)\nWARNING: %s found %s's log file had too many entries, data discarded\n",program,router);
		    break;
		}
		hist = lh + (hist - history);
		history = lh;
	    }
	    hist++;
	}
	histvalid = x;
	fclose (fi);
    } else {
	retcode = 1;
    }
    return (retcode);
}

int
readfile ()
{
    char buf[128];
    int err, x;
    time_t now;
    struct HISTORY *hist;

    sprintf (buf, "%s.log", router);
    if ((err = readhist (buf)) != 0) {	/* Read of log file failed.  Try backup */
	//fprintf(stderr,"Rateup WARNING: %s could not read the primary log file for %s\n",program,router);
	sprintf (buf, "%s.old", router);
	if ((err = readhist (buf)) != 0) {	/* Backup failed too. New file? */
	    //fprintf(stderr,"Rateup WARNING: %s The backup log file for %s was invalid as well\n",program,router);
	    if (err == 2)
		return (1);

	    /* File does not exist - it must be created */
	    now = NOW - DAY_SAMPLE;
	    hist = history;
	    histvalid = DAY_COUNT + WEEK_COUNT + MONTH_COUNT + YEAR_COUNT - 1;
	    last.time = now;
	    /* calculating a diff does not make sense */
	    last.in[0] = 'x';
	    now /= DAY_SAMPLE;
	    now *= DAY_SAMPLE;
	    for (x = 0; x < DAY_COUNT; x++, hist++) {
		hist->time = now;
		hist->in = hist->inmax = hist->out = hist->outmax = 0;
		now -= DAY_SAMPLE;
	    }
	    now /= WEEK_SAMPLE;
	    now *= WEEK_SAMPLE;
	    for (x = 0; x < WEEK_COUNT; x++, hist++) {
		hist->time = now;
		hist->in = hist->inmax = hist->out = hist->outmax = 0;
		now -= WEEK_SAMPLE;
	    }
	    now /= MONTH_SAMPLE;
	    now *= MONTH_SAMPLE;
	    for (x = 0; x < MONTH_COUNT; x++, hist++) {
		hist->time = now;
		hist->in = hist->inmax = hist->out = hist->outmax = 0;
		now -= MONTH_SAMPLE;
	    }
	    now /= YEAR_SAMPLE;
	    now *= YEAR_SAMPLE;
	    for (x = 0; x < YEAR_COUNT; x++, hist++) {
		hist->time = now;
		hist->in = hist->inmax = hist->out = hist->outmax = 0;
		now -= YEAR_SAMPLE;
	    }
	}
    }
    
    return 0;
    
}

int
update (in_, out_, abs_max, absupdate)
     int in_, out_;
     unsigned long abs_max;
     int absupdate;
{
    FILE *fo;
    char buf[128], buf1[128], buf2[128];
    char in[32], out[32];
    time_t now, nextnow, plannow;
    long inrate, outrate;
    unsigned long avc, inmax, outmax;
    double period, interval;
    int x, n, nout;
    struct HISTORY *lhist, *hist;
    double inr, outr;

    sprintf (in, "%d", in_);
    sprintf (out, "%d", out_);

    now = NOW;

    if (now < last.time) {
	fprintf (stderr, "ERROR: %s found that %s's log file time of %lu was greater than now (%lu)\nERROR: Let's not do the time warp, again. Logfile unchanged.\n", program, router, last.time, now);
	return 1;
    }
    sprintf (buf, "%s.tmp", router);
    sprintf (buf1, "%s.log", router);
    sprintf (buf2, "%s.old", router);
    if ((lhist = calloc (1, sizeof (struct HISTORY) * (MAX_HISTORY + 1))) == NULL) {
	fprintf (stderr, "ERROR: Out of memory in update\n");
	return (1);
    }
    hist = lhist;

    period = (now - last.time);
    if (period <= 0 || period > (60 * 60) || last.in[0] == 'x') {	/* if last update is strange */
	if (unknaszero == 1) {
	    inrate = 0;		/* sync unknown to zero */
	    outrate = 0;
	} else {
	    inrate = history[0].in;	/* Sync by using last value */
	    outrate = history[0].out;
	}
    } else {
	/* gauge and absolute */
	if ((absupdate != 0) && (absupdate != 3) && (absupdate != 4)) {
	    strcpy (last.in, "0");
	    strcpy (last.out, "0");
	}
	/* gauge option */
	if (absupdate == 2) {
	    inrate = diff (in, last.in);
	    outrate = diff (out, last.out);
	} else if (absupdate == 3) {
	    inrate = diff (in, last.in) * (3600.0 / (period * 1.0));
	    outrate = diff (out, last.out) * (3600.0 / (period * 1.0));
	} else if (absupdate == 4) {
	    inrate = diff (in, last.in) * (60.0 / (period * 1.0));
	    outrate = diff (out, last.out) * (60.0 / (period * 1.0));
	} else {
	    inrate = diff (in, last.in) / period;
	    outrate = diff (out, last.out) / period;
	}
    }
    if (inrate < 0 || inrate > abs_max) {
	if (unknaszero == 1) {
	    inrate = 0;		/* sync unknown to zero */
	} else {
	    inrate = history[0].in;	/* Sync by using last value */
	}
    }
    if (outrate < 0 || outrate > abs_max) {
	if (unknaszero == 1) {
	    outrate = 0;	/* sync unknown to zero */
	} else {
	    outrate = history[0].out;	/* Sync by using last value */
	}
    }
    if ((fo = fopen (buf, "w")) != NULL) {
	fprintf (fo, "%lu %s %s\n", now, in, out);
	last.time = now;
	/* what is this good for? */
	/* gauge und absolute */
	if ((absupdate != 0) && (absupdate != 3) && (absupdate != 4)) {
	    strcpy (last.in, "0");
	    strcpy (last.out, "0");
	} else {
	    strcpy (last.in, in);
	    strcpy (last.out, out);
	}
	fprintf (fo, "%lu %lu %lu %lu %lu\n", now, inrate, outrate, inrate, outrate);
	nout = 1;
	hist->time = now;
	hist->in = inrate;
	hist->out = outrate;
	hist->inmax = inrate;
	hist->outmax = outrate;
	hist++;

	/* just in case we were dead for a long time, don't try to gather 
	   data from non existing log entries  */
	now = plannow = history[0].time;

	plannow /= DAY_SAMPLE;
	plannow *= DAY_SAMPLE;
	n = 0;

	/* gobble up every shred of data we can get ... */
	if (plannow < now) {
	    NEXT ((now - plannow));
	    fprintf (fo, "%lu %lu %lu %lu %lu\n", now,
		  (unsigned long) inr, (unsigned long) outr, inmax, outmax);
	    hist->time = now;
	    hist->in = inr;
	    hist->out = outr;
	    hist->inmax = inmax;
	    hist->outmax = outmax;
	    nout++;
	    hist++;
	    now = plannow;

	}
	for (x = 1; x < DAY_COUNT; x++) {
	    NEXT (DAY_SAMPLE);
	    fprintf (fo, "%lu %lu %lu %lu %lu\n", now,
		  (unsigned long) inr, (unsigned long) outr, inmax, outmax);
	    hist->time = now;
	    hist->in = inr;
	    hist->out = outr;
	    hist->inmax = inmax;
	    hist->outmax = outmax;
	    nout++;
	    hist++;
	    now -= DAY_SAMPLE;
	}

	plannow = now;
	plannow /= WEEK_SAMPLE;
	plannow *= WEEK_SAMPLE;

	if (plannow < now) {
	    NEXT ((now - plannow));
	    fprintf (fo, "%lu %lu %lu %lu %lu\n", now,
		  (unsigned long) inr, (unsigned long) outr, inmax, outmax);
	    hist->time = now;
	    hist->in = inr;
	    hist->out = outr;
	    hist->inmax = inmax;
	    hist->outmax = outmax;
	    nout++;
	    hist++;
	    now = plannow;

	}
	for (x = 0; x < WEEK_COUNT; x++) {
	    NEXT (WEEK_SAMPLE);
	    fprintf (fo, "%lu %lu %lu %lu %lu\n", now,
		  (unsigned long) inr, (unsigned long) outr, inmax, outmax);
	    hist->time = now;
	    hist->in = inr;
	    hist->out = outr;
	    hist->inmax = inmax;
	    hist->outmax = outmax;
	    nout++;
	    hist++;
	    now -= WEEK_SAMPLE;
	}

	plannow = now;
	plannow /= MONTH_SAMPLE;
	plannow *= MONTH_SAMPLE;

	if (plannow < now) {
	    NEXT ((now - plannow));
	    fprintf (fo, "%lu %lu %lu %lu %lu\n", now,
		  (unsigned long) inr, (unsigned long) outr, inmax, outmax);
	    hist->time = now;
	    hist->in = inr;
	    hist->out = outr;
	    hist->inmax = inmax;
	    hist->outmax = outmax;
	    nout++;
	    hist++;
	    now = plannow;

	}
	for (x = 0; x < MONTH_COUNT; x++) {
	    NEXT (MONTH_SAMPLE);
	    fprintf (fo, "%lu %lu %lu %lu %lu\n", now,
		  (unsigned long) inr, (unsigned long) outr, inmax, outmax);
	    hist->time = now;
	    hist->in = inr;
	    hist->out = outr;
	    hist->inmax = inmax;
	    hist->outmax = outmax;
	    nout++;
	    hist++;
	    now -= MONTH_SAMPLE;
	}

	plannow = now;
	plannow /= YEAR_SAMPLE;
	plannow *= YEAR_SAMPLE;

	if (plannow < now) {
	    NEXT ((now - plannow));
	    fprintf (fo, "%lu %lu %lu %lu %lu\n", now,
		  (unsigned long) inr, (unsigned long) outr, inmax, outmax);
	    hist->time = now;
	    hist->in = inr;
	    hist->out = outr;
	    hist->inmax = inmax;
	    hist->outmax = outmax;
	    nout++;
	    hist++;
	    now = plannow;

	}
	for (x = 0; x < YEAR_COUNT; x++) {
	    NEXT (YEAR_SAMPLE);
	    fprintf (fo, "%lu %lu %lu %lu %lu\n", now,
		  (unsigned long) inr, (unsigned long) outr, inmax, outmax);
	    hist->time = now;
	    hist->in = inr;
	    hist->out = outr;
	    hist->inmax = inmax;
	    hist->outmax = outmax;
	    nout++;
	    hist++;
	    now -= YEAR_SAMPLE;
	}

	if (ferror (fo) || fclose (fo)) {
	    perror (program);
	    fprintf (stderr, "ERROR: Can't write new log file\n");
	    return (1);
	}
	/* another fix to get things working under NT */
	if (unlink (buf2)) {
	    //fprintf(stderr,"Rateup WARNING: %s Can't remove %s updating log file\n",              program,buf2);
	}
	if (rename (buf1, buf2)) {
	    //fprintf(stderr,"Rateup WARNING: %s Can't rename %s to %s updating log file\n",                program,buf1,buf2);
	}
	if (rename (buf, buf1)) {
	    //fprintf(stderr,"Rateup WARNING: %s Can't rename %s to %s updating log file\n",                program,buf,buf1);
	}
	for (n = 0; n < nout && n < MAX_HISTORY; n++) {
	    history[n] = lhist[n];
	}
    } else {
	perror (program);
	fprintf (stderr, "ERROR: Can't open %s for write\n", buf);
	return (1);
    }
    free (lhist);
    
    return 0;
        
}

void
init_colour (int *colmap, int c0, int c1, int c2)
{
    *colmap++ = c0;
    *colmap++ = c1;
    *colmap = c2;
}

struct image_descr id_5min = {
	460, 100,    /* maxx , maxy */
	1.0, 1.0,    /* xscale , yscale */
	0, 300, 0,   /* growright, step, bits */   
	4, 1.0,      /* ytics , yticsf */
	0,           /* peak */

	"MAXIN", 
	"AVGIN", 
	"CURIN", 
	"MAXOUT",
	"AVGOUT",	
	"CUROUT",	
};
	
struct image_descr id_30min = {
	460, 100,    /* maxx , maxy */
	1.0, 1.0,    /* xscale , yscale */
	0, 1800, 0,  /* growright, step, bits */   
	4, 1.0,      /* ytics , yticsf */
	0,           /* peak */

	"HMAXIN", 
	"HAVGIN", 
	"HCURIN", 
	"HMAXOUT",
	"HAVGOUT",	
	"HCUROUT",	
};

struct image_descr id_120min = {
	460, 100,    /* maxx , maxy */
	1.0, 1.0,    /* xscale , yscale */
	0, 7200, 0,  /* growright, step, bits */   
	4, 1.0,      /* ytics , yticsf */
	0,           /* peak */

	"TMAXIN", 
	"TAVGIN", 
	"TCURIN", 
	"TMAXOUT",
	"TAVGOUT",	
	"TCUROUT",	
};

struct image_descr id_1440min = {
	460, 100,    /* maxx , maxy */
	1.0, 1.0,    /* xscale , yscale */
	0, 86400, 0, /* growright, step, bits */   
	4, 1.0,      /* ytics , yticsf */
	0,           /* peak */

	"DMAXIN", 
	"DAVGIN", 
	"DCURIN", 
	"DMAXOUT",
	"DAVGOUT",	
	"DCUROUT",	
};


char * 
mkfilename(char *dir, char *name, char *suffix)
{
	int ret;
	char *buf = malloc(strlen(dir) + 1
			   + strlen(name)
			   + strlen(suffix) + 1);
	if (buf)
		sprintf(buf, "%s/%s%s", dir, name, suffix);
	return buf;
}

int
frateup (char *realdir_,
         char *rname_,
         char *ip_,
         char *imgf_,
         int bin_,
         int bout_,
         int bmax_)
{
	pp_tab_t *tab = NULL;
	char *p;
	char *tabfile;
	struct tm *t_;
	time_t tnow_;
	int hour_, minutes_;
	int init;
	routerpath = realdir_;
    
	host_ip = ip_;
	host_speed = bmax_;
    
	bmax_ /= 8;

	if ( chdir (routerpath) ) {
		fprintf (stderr, "chdir to %s failed\n", routerpath);
 		return 1;
	}
	
	/* Initialiase the colour variables  - should be overwritten */
	init_colour (&col_in[0], c_in);
	init_colour (&col_out[0], c_out);
	init_colour (&col_inm[0], c_inm);
	init_colour (&col_outm[0], c_outm);
	init_colour (&col_outp[0], c_outp);

	if ((history = calloc (1, sizeof (struct HISTORY) * (MAX_HISTORY + 1))) == NULL) {
		fprintf (stderr, "out of memory\n");
		return 1;
	}

	Mh = MAX_HISTORY;
	router = rname_;
	NOW = time (NULL);

	readfile ();
	tabfile = mkfilename(routerpath, router, ".tab");
	init = read_symtab(&tab, tabfile);	

	transparent = 1;
	unknaszero = 1;
	dorelpercent = 1;

	update (bin_/id_5min.step, bout_/id_5min.step, bmax_, 2);

	col_in[0] = 0x00;
	col_in[1] = 0xcc;
	col_in[2] = 0x00;
	col_out[0] = 0x00;
	col_out[1] = 0x00;
	col_out[2] = 0xff;
	col_inm[0] = 0x00;
	col_inm[1] = 0x66;
	col_inm[2] = 0x00;
	col_outm[0] = 0xff;
	col_outm[1] = 0x00;
	col_outm[2] = 0xff;

	/* 5 min avg graphs */
	image (imgf_, bmax_, bmax_, &id_5min, &tab);

	/* there should be 30, 120, and 24*60 min avgs */
	
	/* see if we have to update weekly, monthly or yearly graph */
	tnow_ = time ( &tnow_ );
	t_ = localtime ( &tnow_ );
	hour_ = t_->tm_hour;	
	minutes_ =  t_->tm_min;
	
	/* first update weekly graph ( every 30 minutes ) */
	p = mkfilename(routerpath, router, "-week.png");
	if (init || access(p, R_OK)
	    || (minutes_ >= 0 && minutes_ < 5)
	    || (minutes_ >= 30 && minutes_ < 35)) 
		image (p, bmax_, bmax_, &id_30min, &tab);
	free(p);
	
	/* next update monthly graph ( every 2 hours ) */
	p = mkfilename(routerpath, router, "-month.png");
	if (init || access(p, R_OK)
	    || (minutes_ >= 0 && minutes_ < 5 && (hour_ % 2) == 0))
		image (p, bmax_, bmax_, &id_120min, &tab);
	free(p);

	/* and finally - yearly graph ( once per day ) */
	p = mkfilename(routerpath, router, "-year.png");
	if (init || access(p, R_OK)
	    || (minutes_ >= 0 && minutes_ < 5 && hour_ == 0))
		image (p, bmax_, bmax_, &id_1440min, &tab);
	free(p);

	/* Override values of certain variables. This is based on
	   the fact that add_.*() calls append the symbol to the
	   beginning of the symbol table */ 
	add_string_value(&tab, "PROGRAM", PACKAGE_NAME);
	add_string_value(&tab, "VERSION", PACKAGE_VERSION);
	add_string_value(&tab, "ROUTERNAME", router);
	add_string_value(&tab, "ROUTERIP", host_ip);
	add_string_value(&tab, "NOW", ctime (&NOW));
	
	add_numeric_value(&tab, "SPEED", host_speed);

	/* Create html page */
	sprintf (htmlf, "%s/%s.html", routerpath, router);
	create_html(tab, html_template, htmlf);
	write_symtab(tab, tabfile);
	free(tabfile);
	free_tab(&tab);

	return 0;
}
